"""
The state module.

Stores attributes for the state instance and handles device-related
"""
import logging
from ..connection import message_data
from ..errors import wappsto_errors


class State:
    """
    State instance class.

    Stores attributes for the state instance and handles device-related
    """

    def __init__(self, parent, uuid, state_type, timestamp, init_value):
        """
        Initialize the State class.

        Initializes an object of network class by passing required parameters.

        Args:
            parent: reference to a value object to which a state
                belongs to
            uuid: unique identifier of a state
            state_type: determines if the state is report or control
            timestamp: time of last update
            init_value: Initial value after creation of an object

        """
        self.wapp_log = logging.getLogger(__name__)
        self.wapp_log.addHandler(logging.NullHandler())
        self.parent = parent
        self.uuid = uuid
        self.state_type = state_type
        self.timestamp = timestamp
        self.callback = None

        self.init_value = init_value
        self.data = init_value

        msg = "State {} Debug: \n{}".format(uuid, str(self.__dict__))
        self.wapp_log.debug(msg)

    def get_parent_value(self):  # pragma: no cover
        """
        Retrieve parent value reference.

        Gets a reference to the value that owns this state.

        Returns:
            Reference to instance of value class that owns this state.

        """
        return self.parent

    def set_callback(self, callback):
        """
        Set the callback.

        Sets the callback attribute.

        Args:
            callback: Callback reference.

        Raises:
            CallbackNotCallableException: Custom exception to signify invalid
            callback.

        """
        if not callable(callback):
            msg = "Callback method should be a method"
            self.wapp_log.error("Error setting callback: {}".format(msg))
            raise wappsto_errors.CallbackNotCallableException
        self.callback = callback
        self.wapp_log.debug("Callback {} has been set.".format(callback))
        return True

    def handle_delete(self):
        """
        Handle delete.

        Calls the __call_callback method with initial input of "remove".

        Returns:
            result of __call_callback method.

        """
        self.__call_callback('remove')

    def delete(self):
        """
        Delete this object.

        Sends delete request for this object and removes its reference
        from parent.

        """
        message = message_data.MessageData(
            message_data.SEND_DELETE,
            network_id=self.parent.parent.parent.uuid,
            device_id=self.parent.parent.uuid,
            value_id=self.parent.uuid,
            state_id=self.uuid
        )
        self.parent.parent.parent.conn.sending_queue.put(message)
        if self == self.parent.report_state:
            self.parent.report_state = None
            self.wapp_log.info("Report state removed")
        elif self == self.parent.control_state:
            self.parent.control_state = None
            self.wapp_log.info("Control state removed")

    def __call_callback(self, event):
        if self.callback is not None:
            self.callback(self, event)



"""
The receive_data module.

Handles incoming data from the server.

"""
import sys
import json
import random
import logging
from . import message_data
from json.decoder import JSONDecodeError

RECEIVE_SIZE = 1024
MESSAGE_SIZE_BYTES = 1000000


class ReceiveData:
    """The ReceiveData class that handles receiving information."""

    def __init__(self, client_socket):
        """
        Initialize the ReceiveData class.

        Initializes an object of ReceiveData class by passing required
        parameters. While initialization, wapp_log is created.

        Args:
            client_socket: reference to ClientSocket instance.

        """
        self.wapp_log = logging.getLogger(__name__)
        self.wapp_log.addHandler(logging.NullHandler())

        self.client_socket = client_socket

    def __get_random_id(self):
        network_n = self.client_socket.data_manager.network.name
        random_int = random.randint(1, 25000)
        return "{}{}".format(network_n, random_int)

    def sending_queue_add_trace(self, parent, trace_id, data, control_value_id=None):
        """
        Add a trace to the sending queue.

        Adds a trace URL to the sending queue for debugging purposes.

        Args:
            parent: Owner of the trace URL.
            trace_id: ID of the trace message.
            data: Trace message data.
            control_value_id: ID of the control state of the value
                (default: {None})

        """
        if trace_id:
            trace = message_data.MessageData(
                message_data.SEND_TRACE,
                parent=parent,
                trace_id=trace_id,
                data=data,
                text="ok",
                control_value_id=control_value_id)
            self.client_socket.sending_queue.put(trace)

    def receive_thread(self):
        """
        Create the receive thread.

        Starts a while True loop checking if something is received.
        """
        self.wapp_log.debug("ReceiveThread Started!")
        while True:
            self.receive_message()

    def receive_data(self):
        """
        Socket receive method.

        Method that handles receiving data from a socket. Capable of handling
        data chunks.

        Returns:
            The decoded message from the socket.

        """
        total_decoded = ''
        decoded = None
        while True:
            if self.client_socket.connected:
                data = self.client_socket.my_socket.recv(RECEIVE_SIZE)
                if data == b'':
                    self.client_socket.reconnect()
                    return None
                try:
                    decoded_data = data.decode('utf-8')
                except AttributeError:
                    continue
                total_decoded += decoded_data
                if sys.getsizeof(total_decoded) > MESSAGE_SIZE_BYTES:
                    error = "Received message exeeds size limit."
                    self.wapp_log.error(error)
                    return None
                try:
                    decoded = json.loads(total_decoded)
                except JSONDecodeError:
                    if len(decoded_data) < RECEIVE_SIZE:
                        error = "Json decoding error: {}".format(total_decoded)
                        self.wapp_log.error(error)
                        return None
                else:
                    break
            else:
                break
        return decoded

    def receive_message(self):
        """
        Receives message.

        Receives message and passes it to receive method, and catches
        encountered exceptions.
        """
        try:
            decoded = self.receive_data()

            # if the received string is list
            if isinstance(decoded, list):
                for decoded_data in decoded:
                    self.receive(decoded_data)
            else:
                self.receive(decoded)

        except ConnectionResetError as e:  # pragma: no cover
            msg = "Received Reset: {}".format(e)
            self.wapp_log.error(msg, exc_info=True)
            self.client_socket.reconnect()

        except OSError as oe:  # pragma: no cover
            msg = "Received OS Error: {}".format(oe)
            self.wapp_log.error(msg, exc_info=True)
            self.client_socket.reconnect()

    def receive(self, decoded):
        """
        Performs acction on received message.

        Based on the type of message, directs the decoded data to the
        appropriate methods.

        Args:
            decoded: the received message

        """
        if decoded:
            try:
                self.wapp_log.debug('Raw received Json: {}'.format(decoded))
                if decoded.get('method', False) == 'PUT':
                    self.incoming_put(decoded)

                elif decoded.get('method', False) == 'GET':
                    self.incoming_get(decoded)

                elif decoded.get('method', False) == 'DELETE':
                    self.incoming_delete(decoded)

                elif decoded.get('error', False):
                    self.incoming_error(decoded)

                elif decoded.get('result', False):
                    self.incoming_result(decoded)

                else:
                    return_id = decoded.get('id')
                    self.wapp_log.warning("Unhandled method")
                    error_str = 'Unknown method'
                    self.error_reply(error_str, return_id)

            except ValueError:
                return_id = decoded.get('id')
                error_str = 'Value error'
                self.wapp_log.error("{} [{}]: {}".format(error_str, return_id, decoded))
                self.error_reply(error_str, return_id)

    def incoming_put(self, data):
        """
        Incoming data handler.

        Sends the data from incoming control messages to the appropriate
        handler method.

        Args:
            data: JSON communication message data.

        Returns:
            Results of the incoming control handling.

        """
        return_id = data.get('id')
        try:
            uuid = data.get('params').get('data').get('meta').get('id')
            meta_type = data.get('params').get('data').get('meta').get('type')
            self.wapp_log.debug("Put request from id: " + uuid)
        except AttributeError:
            error_str = 'Error received incorrect format in put: {}'.format(str(data))
            self.wapp_log.error(error_str, exc_info=True)
            return

        try:
            trace_id = data.get('params').get('meta').get('trace')
            if trace_id:
                self.wapp_log.debug("Found trace id: " + trace_id)
        except AttributeError:
            trace_id = None

        obj = self.client_socket.data_manager.get_by_id(uuid)
        if obj is None:
            self.error_reply('Non-existing uuid provided', return_id)
            return

        param_data = data.get('params').get('data')
        try:
            if meta_type == "value":
                period = param_data.get('period')
                if period:
                    obj.set_period(period)
                delta = param_data.get('delta')
                if delta:
                    obj.set_delta(delta)
                self.success_reply(return_id)
                self.sending_queue_add_trace(
                    obj.parent.uuid,
                    trace_id,
                    None,
                    control_value_id=self.__get_random_id()
                )
            elif meta_type == "state":
                local_data = param_data.get('data')
                if obj.state_type == "Control":
                    self.success_reply(return_id)
                    obj.parent.handle_control(data_value=local_data)
                    self.sending_queue_add_trace(
                        obj.parent.uuid,
                        trace_id,
                        local_data,
                        control_value_id=self.__get_random_id()
                    )
                else:
                    self.error_reply('Element is not control state', return_id)
        except AttributeError:
            self.error_reply('Attribute error encountered', return_id)

    def incoming_get(self, data):
        """
        Incoming data handler.

        Sends the data from incoming report messages to the appropriate handler
        method.

        Args:
            data: JSON communication message data.

        Returns:
            Results of the incoming report handling.

        """
        return_id = data.get('id')
        try:
            uuid = data.get('params').get('url').split('/')[-1]
            self.wapp_log.debug("Get request from id: " + uuid)
        except AttributeError:
            error_str = 'Error received incorrect format in get: {}'.format(str(data))
            self.wapp_log.error(error_str, exc_info=True)
            return

        try:
            trace_id = data.get('params').get('meta').get('trace')
            if trace_id:
                self.wapp_log.debug("Found trace id: " + trace_id)
        except AttributeError:
            trace_id = None

        obj = self.client_socket.data_manager.get_by_id(uuid)
        if obj is None:
            self.error_reply('Non-existing uuid provided', return_id)
            return

        try:
            if obj.state_type == "Report":
                self.success_reply(return_id)
                self.sending_queue_add_trace(
                    obj.parent.uuid,
                    trace_id,
                    obj.data,
                    control_value_id=self.__get_random_id()
                )
                obj.parent.handle_refresh()
            else:
                self.error_reply('Element is not control state', return_id)
        except AttributeError:
            self.error_reply('Attribute error encountered', return_id)

    def incoming_delete(self, data):
        """
        Incoming delete handler.

        Sends the event from incoming delete messages to the
        appropriate handler method.

        Args:
            data: JSON communication message data.

        Returns:
            Results of the incoming report handling.

        """
        return_id = data.get('id')
        try:
            uuid = data.get('params').get('url').split('/')[-1]
            self.wapp_log.debug("Delete request from id: " + uuid)
        except AttributeError:
            error_str = 'Error received incorrect format in delete: {}'.format(str(data))
            self.wapp_log.error(error_str, exc_info=True)
            return

        try:
            trace_id = data.get('params').get('meta').get('trace')
            if trace_id:
                self.wapp_log.debug("Found trace id: " + trace_id)
        except AttributeError:
            trace_id = None

        obj = self.client_socket.data_manager.get_by_id(uuid)
        if obj is None:
            self.error_reply('Non-existing uuid provided', return_id)
            return

        try:
            self.success_reply(return_id)
            self.sending_queue_add_trace(
                obj.uuid,
                trace_id,
                None,
                control_value_id=self.__get_random_id()
            )
            obj.handle_delete()
        except AttributeError:
            self.error_reply('Attribute error encountered', return_id)

    def incoming_error(self, data):
        """
        Incoming error handler.

        Deals with incoming error message.

        Args:
            data: JSON communication message data.

        """
        return_id = data.get('id')
        msg = "Error: {}".format(data.get('error').get('message'))
        self.wapp_log.error(msg)
        self.client_socket.remove_id_from_confirm_list(return_id)

    def incoming_result(self, data):
        """
        Incoming result handler.

        Deals with incoming result message.

        Args:
            data: JSON communication message data.

        """
        return_id = data.get('id')
        result_value = data['result'].get('value', True)
        if result_value is not True:
            uuid = result_value['meta']['id']
            data = result_value['data']
            object = self.client_socket.data_manager.get_by_id(uuid)
            if object is not None and object.parent.control_state == object:
                object.parent.handle_control(data_value=data)
        self.client_socket.remove_id_from_confirm_list(return_id)

    def success_reply(self, return_id):
        """
        Handle successful replies on the receive thread.

        Adds success message to the sending_queue.

        Args:
            return_id: ID of the success message.

        """
        success_reply = message_data.MessageData(
            message_data.SEND_SUCCESS,
            rpc_id=return_id
        )
        self.client_socket.sending_queue.put(success_reply)

    def error_reply(self, error_str, return_id):
        """
        Handle error replies on the receive thread.

        Adds error message to the sending_queue.

        Args:
            error_str: Error message contents.
            return_id: ID of the error message.

        """
        error_reply = message_data.MessageData(
            message_data.SEND_FAILED,
            rpc_id=return_id,
            text=error_str
        )
        self.client_socket.sending_queue.put(error_reply)



"""
The device module.

Stores attributes for the device instance and handles device-related
methods.
"""
import logging
import warnings
from ..connection import message_data
from ..errors import wappsto_errors


class Device:
    """
    Device instance class.

    Stores attributes for the device instance and handles device-related
    methods.
    """

    def __init__(
        self,
        parent,
        uuid,
        name,
        product,
        protocol,
        serial_number,
        version,
        manufacturer,
        communication,
        description
    ):
        """
        Initialize the Device class.

        Initializes an object of device class by passing required parameters.

        Args:
            parent: Reference to an object of network class
            uuid: Unique identifier of a device
            name: Name of a device
            product: Defines what a product is
            protocol: Protocol device communicates by
            serial_number: Serial number of a device
            version: Version of a device
            manufacturer: Defines who made a product
            communication: The way a device communicates
            description: Description of a device

        """
        self.wapp_log = logging.getLogger(__name__)
        self.wapp_log.addHandler(logging.NullHandler())
        self.parent = parent
        self.uuid = uuid
        self.name = name
        self.product = product
        self.serial_number = serial_number
        self.version = version
        self.manufacturer = manufacturer
        self.communication = communication
        self.protocol = protocol
        self.description = description
        self.values = []
        self.callback = None
        msg = "Device {} Debug: \n {}".format(name, str(self.__dict__))
        self.wapp_log.debug(msg)

    def __getattr__(self, attr):  # pragma: no cover
        """
        Get attribute value.

        When trying to get value from value_list warning is raised about
        it being deprecated and calls values instead.

        Returns:
            values

        """
        if attr in ["value_list"]:
            warnings.warn("Property %s is deprecated" % attr)
            return self.values

    def get_parent_network(self):  # pragma: no cover
        """
        Retrieve parent network reference.

        Gets a reference to the network that owns this device.

        Returns:
            Reference to instance of Network class that owns this Device.

        """
        return self.parent

    def add_value(self, value):
        """
        Append value reference to the value list.

        Adds a value reference to the list of value references
        ("owned" values).

        Args:
            value: Reference to instance of Value class.

        """
        self.values.append(value)
        self.wapp_log.debug("Value {} has been added.".format(value))

    def get_value(self, value_name):
        """
        Retrieve child value reference.

        Gets a child Value class reference by the given name string.

        Args:
            value_name: String to search by in the list of values.

        Returns:
            Reference to instance of Value class.

        """
        for value in self.values:
            if value_name == value.name:
                return value
        else:
            msg = "Value {} not found".format(value_name)
            raise wappsto_errors.ValueNotFoundException(msg)

    def set_callback(self, callback):
        """
        Set the callback.

        Sets the callback attribute.

        Args:
            callback: Callback reference.

        Raises:
            CallbackNotCallableException: Custom exception to signify invalid
            callback.

        """
        if not callable(callback):
            msg = "Callback method should be a method"
            self.wapp_log.error("Error setting callback: {}".format(msg))
            raise wappsto_errors.CallbackNotCallableException
        self.callback = callback
        self.wapp_log.debug("Callback {} has been set.".format(callback))
        return True

    def handle_delete(self):
        """
        Handle delete.

        Calls the __call_callback method with initial input of "remove".

        Returns:
            result of __call_callback method.

        """
        self.__call_callback('remove')

    def delete(self):
        """
        Delete this object.

        Sends delete request for this object and removes its reference
        from parent.

        """
        message = message_data.MessageData(
            message_data.SEND_DELETE,
            network_id=self.parent.uuid,
            device_id=self.uuid
        )
        self.parent.conn.sending_queue.put(message)
        self.parent.devices.remove(self)
        self.wapp_log.info("Device removed")

    def __call_callback(self, event):
        if self.callback is not None:
            self.callback(self, event)



"""
The wappsto encoding module.

Handles encoding object instances to a JSON file.
"""
import logging
from ..connection.seluxit_rpc import SeluxitRpc


class WappstoEncoder:
    """
    The wappsto encoding class.

    Handles encoding the current runtime object instances into JSON. This
    allows the system to be saved as a parsable JSON file similar to the one
    used to start the package.
    """

    def __init__(self):
        """
        Initialize WappstoEncoder.

        Initializes the WappstoEncoder class.
        """
        self.wapp_log = logging.getLogger(__name__)
        self.wapp_log.addHandler(logging.NullHandler())

    def encode_network(self, network):
        """
        Encode instance of Network class.

        Handles the encoding of the network instance, contains a template to
        encode the network with.

        Args:
            network: Reference to the instance of the Network class.

        Returns:
            The dictionary.

        """
        encoded_devices = []
        for device in network.devices:
            encoded_device = self.encode_device(device)
            encoded_devices.append(encoded_device)

        encoded_network = {
            'name': network.name,
            'device': encoded_devices,
            'meta': {
                'id': network.uuid,
                'version': '2.0',
                'type': 'network'
            }
        }

        if SeluxitRpc.is_upgradable():
            encoded_network.get('meta').update({'upgradable': True})

        self.wapp_log.debug("Network JSON: {}".format(encoded_network))
        return encoded_network

    def encode_device(self, device):
        """
        Encode instance of Device class.

        Handles the encoding of the device instance, contains a template to
        encode the device with.

        Args:
            device: Reference to the instance of the Device class.

        Returns:
            The dictionary.

        """
        encoded_values = []
        for value in device.values:
            encoded_value = self.encode_value(value)
            encoded_values.append(encoded_value)

        encoded_device = {
            'name': device.name,
            'product': device.product,
            'protocol': device.protocol,
            'serial': device.serial_number,
            'manufacturer': device.manufacturer,
            'communication': device.communication,
            'description': device.description,
            'version': device.version,
            'value': encoded_values,
            'meta':
            {
                'id': device.uuid,
                'version': '2.0',
                'type': 'device'
            }
        }

        self.wapp_log.debug("Device JSON: {}".format(encoded_device))
        return encoded_device

    def encode_value(self, value):
        """
        Encode instance of Value class.

        Handles the encoding of the value instance, contains a template to
        encode the value with.

        Args:
            value: Reference to the instance of the Value class.

        Returns:
            The dictionary.

        """
        states = []
        if value.report_state:
            encoded_state = self.encode_state(
                value.report_state
            )
            states.append(encoded_state)

        if value.control_state:
            encoded_state = self.encode_state(
                value.control_state
            )
            states.append(encoded_state)

        if value.data_type == 'string':
            details = {
                'encoding': value.string_encoding,
                'max': value.string_max
            }
        elif value.data_type == 'blob':
            details = {
                'encoding': value.blob_encoding,
                'max': value.blob_max
            }
        elif value.data_type == 'number':
            details = {
                'min': value.number_min,
                'max': value.number_max,
                'step': value.number_step,
                'unit': value.number_unit
            }

        encoded_value = {
            'name': value.name,
            'type': value.type_of_value,
            'permission': value.permission,
            'state': states,
            value.data_type: details,
            'meta':
            {
                'id': value.uuid,
                'type': 'value',
                'version': '2.0'
            }
        }

        self.wapp_log.debug("Value JSON: {}".format(encoded_value))
        return encoded_value

    def encode_state(self, state):
        """
        Encode instance of State class.

        Handles the encoding of the state instance, contains a template to
        encode the state with.

        Args:
            state: Reference to the instance of the State class.

        Returns:
            The dictionary.

        """
        encoded_state = {
            'data': state.data,
            'type': state.state_type,
            'timestamp': state.timestamp,
            'meta':
            {
                'id': state.uuid,
                'type': 'state',
                'version': '2.0',
                'contract': []
            }
        }

        self.wapp_log.debug("State JSON: {}".format(encoded_state))
        return encoded_state
